// Copyright 2017 The Johns Hopkins University Applied Physics Laboratory.
// Licensed under the MIT License. See LICENSE.txt in the project root for full license information.
#include <iostream>
#include <fstream>
#include <sstream>
#include <vector>

#include <gdal.h>
#include <gdal_priv.h>
#include <gdal_alg.h>

#include <ogr_spatialref.h>

#include <tclap/CmdLine.h>

#include <eopcc/utility.hpp>
#include <eopcc/cloud_io.hpp>
#include <eopcc/gdal_io.hpp>

#if defined(WIN32) && (_MSC_VER < 1800)
int round(double val) {
	return (val < 0) ? ceil(val - 0.5) : floor(val + 0.5);
}
#else
using std::round;
#endif

bool FillIntensityFromImage(eopcc::PointCloud& point_cloud, eopcc::ImageType& img, const eopcc::CropInfo& crop_info, void* transformer,
							double min_intensity, double intensity_scale) {
	OGRSpatialReference pj_utm(point_cloud.wkt_spatial_ref.c_str());
	OGRSpatialReference pj_latlong;
	pj_latlong.importFromProj4("+proj=latlong +datum=WGS84");

	OGRCoordinateTransformation* coordinate_transform;
	coordinate_transform = OGRCreateCoordinateTransformation(&pj_utm, &pj_latlong);


 	for (size_t i=0; i < point_cloud.size(); ++i) {
 		double lon = point_cloud.loc[i].x;
 		double lat = point_cloud.loc[i].y;
 		double z = point_cloud.loc[i].z;

 		coordinate_transform->Transform(1, &lon, &lat);



		double pixel_i, pixel_j;
		pixel_i = lon;
		pixel_j = lat;
		int success;
		GDALRPCTransform(transformer, true, 1, &pixel_i, &pixel_j, &z, &success);
		if (!success)
			continue;

		int ii, jj;

		ii = round(pixel_i) - crop_info.pixel_begin;
		jj = round(pixel_j) - crop_info.line_begin;


		//std::cout << point_cloud.loc[i] - point_cloud.extent.min_corner() << ": " << ii << ", " << jj << std::endl;

		if (ii < 0 || jj < 0)
			continue;
		if (ii >= img.width || jj >= img.height)
			continue;


		double intensity = 0.0;
		for (int ik = 0; ik < img.bands; ik++) {
			int data = img.data[jj*img.width + ii*img.bands+ik];
			intensity = intensity + data;
		}

		intensity = intensity/img.bands;
		intensity = (intensity - min_intensity)*intensity_scale;

		point_cloud.intensity[i] = intensity;
 	}

 	return true;
}

void GetImageSpaceExtent(const eopcc::PolygonExtent& poly_extent, void* latlon_to_image,
												 double& min_pixel, double& max_pixel, double& min_line, double& max_line) {
	double average_z = const_cast<eopcc::PolygonExtent&>(poly_extent).average_z();
	double radius = 1000;
	double long_to_pixel=0;
	double lat_to_line=0;
	int success;
	for (auto line=poly_extent.sides.begin(); line != poly_extent.sides.end(); ++line) {
		//centrod
		long_to_pixel += 0.5*(line->start.x+line->end.x);//line->min_corner().x;
		lat_to_line += 0.5*(line->start.y+line->end.y);//line->min_corner().y;
		double z = average_z;
		

		//printf("0line->start.x:%lf line->end.x:%lf line->start.y:%lf line->end.y:%lf\n",line->start.x,line->end.x,line->start.y,line->end.y);

		//double long_to_pixel = line->start.x;//line->min_corner().x;
		//double lat_to_line = line->start.y;//line->min_corner().y;
		//double z = average_z;
		//int success;

		//GDALRPCTransform(latlon_to_image, true, 1, &long_to_pixel, &lat_to_line, &z, &success);

		//if (min_pixel > long_to_pixel) {
		//	min_pixel = long_to_pixel;
		//}
		//if (max_pixel < long_to_pixel) {
		//	max_pixel = long_to_pixel;
		//}

		//if (min_line > lat_to_line) {
		//	min_line = lat_to_line;
		//}
		//if (max_line < lat_to_line) {
		//	max_line = lat_to_line;
		//}
		//printf("1long_to_pixel:%f lat_to_line:%f\n",long_to_pixel,lat_to_line);
		//long_to_pixel = line->end.x;//max_corner().x;
		//lat_to_line = line->end.y;//max_corner().y;
		//z = average_z;

		//GDALRPCTransform(latlon_to_image, true, 1, &long_to_pixel, &lat_to_line, &z, &success);

		//if (min_pixel > long_to_pixel) {
		//	min_pixel = long_to_pixel;
		//}
		//if (max_pixel < long_to_pixel) {
		//	max_pixel = long_to_pixel;
		//}

		//if (min_line > lat_to_line) {
		//	min_line = lat_to_line;
		//}
		//if (max_line < lat_to_line) {
		//	max_line = lat_to_line;
		//}
		//printf("2long_to_pixel:%f lat_to_line:%f\n",long_to_pixel,lat_to_line);
	}
		long_to_pixel/=4;
		lat_to_line/=4;
		GDALRPCTransform(latlon_to_image, true, 1, &long_to_pixel, &lat_to_line, &average_z, &success);

		if (min_pixel > long_to_pixel) {
			min_pixel = long_to_pixel-radius;
		}
		if (max_pixel < long_to_pixel) {
			max_pixel = long_to_pixel+radius;
		}

		if (min_line > lat_to_line) {
			min_line = lat_to_line-radius;
		}
		if (max_line < lat_to_line) {
			max_line = lat_to_line+radius;
		}

		//printf("2long_to_pixel:%f lat_to_line:%f\n",long_to_pixel,lat_to_line);

		
}


// Example program to demonstrate some very basic things you can do with
// to work with NITF images with RPC00B metadata
bool CropImageWithKML(std::string input_file, std::string kml_file, std::string rpc_file, std::string output_file, double spacing) {
	bool ret = true;
	// Read corner coordinates from KML file.
	// OK to set height to zero here to keep things simple.
	eopcc::PolygonExtent poly_extent = eopcc::ExtentFromKML(kml_file);

	// Lat/Long Cube (Really a pyramid extending from the center of the earth).
	eopcc::Extent cube_extent = poly_extent.to_extent();

	std::cout << "Opening "<< input_file << std::endl;
	GDALAllRegister();

	GDALDataset  *poDataset;
	poDataset = (GDALDataset *) GDALOpen( input_file.c_str() , GA_ReadOnly );
    if( poDataset == NULL ) {
	   std::cerr << input_file << " Does Not Exist" << std::endl;
	   return false;
    }

	GDALRasterBand* raster_band = poDataset->GetRasterBand(1);

  int xdim = raster_band->GetXSize();
  int ydim = raster_band->GetYSize();

  std::cout << "Image Dims: " << xdim << ", " << ydim << std::endl;
	std::cout << "  " << std::endl;

	std::cout << "PROJ:"<< poDataset->GetGCPProjection() << std::endl;
	std::cout << "  " << std::endl;

	char **papszMD = NULL;
	GDALRPCInfo sRPC;
	std::cout << "Getting RPC Info" << std::endl;
	char **papszOptions = NULL;
	papszMD = poDataset->GetMetadata( "RPC" );
	GDALExtractRPCInfo(papszMD, &sRPC);

	void* ptr_transform = GDALCreateRPCTransformer(&sRPC, false, 0.1,  nullptr);



	double min_pixel = xdim;
	double max_pixel = 0;
	double min_line = ydim;
	double max_line = 0;

	GetImageSpaceExtent(poly_extent, ptr_transform, min_pixel, max_pixel, min_line, max_line);


	eopcc::CropInfo crop_info;
	eopcc::ImageType img;

	int offset_buffer=0;
	crop_info.pixel_begin = std::max<int>(min_pixel, 0);
	crop_info.pixel_end = std::min<int>(max_pixel, xdim);
	crop_info.line_begin = std::max<int>(min_line, 0);
	crop_info.line_end = std::min<int>(max_line, ydim);

	std::cout << "Crop image  Width: " << crop_info.pixel_begin << " to " << crop_info.pixel_end << std::endl;
	std::cout << "Crop image  Height [from top]: " << crop_info.line_begin << " to " << crop_info.line_end << std::endl;


	img.width = std::abs(crop_info.pixel_end - crop_info.pixel_begin) + 1;
	img.height = std::abs(crop_info.line_end - crop_info.line_begin) + 1;

	if (img.height == 0 || img.width == 0) {
	 	std::cerr << "Image does not overlap with KML borders" << std::endl;
	 	return false;
	}
	int min_intensity = 0;
	int max_intensity = 0;


	char err[1024];
	ret = eopcc::read_cropped_image(input_file.c_str(), img, crop_info, min_intensity, max_intensity, err);
	if (!ret)
		return false;
	std::string extension = output_file.substr(output_file.find_last_of(".") + 1);

	if (extension == "jpeg" || extension == "jpg") {
		ret = eopcc::write_jpeg_file(output_file.c_str(), img, min_intensity, max_intensity, err);
	} else if (extension == "tiff" || extension == "tif") {
		ret = eopcc::write_tiff_file(output_file.c_str(), img, err);
	} else if (extension == "laz" || extension == "las") {
		eopcc::PointCloud sample = eopcc::GenerateSample(kml_file, spacing);
		FillIntensityFromImage(sample, img, crop_info, ptr_transform, min_intensity, 65536.0/(max_intensity - min_intensity));
		eopcc::Write(output_file, sample);

	} else {
		std::cerr << "Unrecognized file extension, defaulting to jpeg" << std::endl;
		ret = eopcc::write_jpeg_file(output_file.c_str(), img, min_intensity, max_intensity, err);
	}

	GDALDestroyTransformer(ptr_transform);
	GDALClose(poDataset);

	if (!ret)
		return false;
    const char* p_rpc_file = rpc_file.c_str();
	FILE *fid=fopen(p_rpc_file,"wb");
	fprintf(fid,"%.20f, %.20f, %.20f, %.20f, %.20f, %.20f, %.20f, %.20f, %.20f, %.20f, %.20f, %.20f, %.20f, %.20f, %.20f, %.20f, %.20f, %.20f, %.20f, %.20f, %.20f, %.20f, %.20f, %.20f, %.20f, %.20f, %.20f, %.20f, %.20f, %.20f, %.20f, %.20f, %.20f, %.20f, %.20f, %.20f, %.20f, %.20f, %.20f, %.20f, %.20f, %.20f, %.20f, %.20f, %.20f, %.20f, %.20f, %.20f, %.20f, %.20f, %.20f, %.20f, %.20f, %.20f, %.20f, %.20f, %.20f, %.20f, %.20f, %.20f, %.20f, %.20f, %.20f, %.20f, %.20f, %.20f, %.20f, %.20f, %.20f, %.20f, %.20f, %.20f, %.20f, %.20f, %.20f, %.20f, %.20f, %.20f, %.20f, %.20f, %.20f, %.20f, %.20f, %.20f, %.20f, %.20f, %.20f, %.20f, %.20f, %.20f, %.20f, %.20f, %.20f, %.20f, %.20f, %.20f\n",
			sRPC.dfLINE_OFF, sRPC.dfSAMP_OFF, sRPC.dfLAT_OFF, sRPC.dfLONG_OFF, sRPC.dfHEIGHT_OFF, sRPC.dfLINE_SCALE, sRPC.dfSAMP_SCALE, sRPC.dfLAT_SCALE, sRPC.dfLONG_SCALE, sRPC.dfHEIGHT_SCALE,
			sRPC.adfLINE_NUM_COEFF[0],sRPC.adfLINE_NUM_COEFF[1],sRPC.adfLINE_NUM_COEFF[2],sRPC.adfLINE_NUM_COEFF[3],sRPC.adfLINE_NUM_COEFF[4],sRPC.adfLINE_NUM_COEFF[5],sRPC.adfLINE_NUM_COEFF[6],sRPC.adfLINE_NUM_COEFF[7],sRPC.adfLINE_NUM_COEFF[8],sRPC.adfLINE_NUM_COEFF[9],
			sRPC.adfLINE_NUM_COEFF[10],sRPC.adfLINE_NUM_COEFF[11],sRPC.adfLINE_NUM_COEFF[12],sRPC.adfLINE_NUM_COEFF[13],sRPC.adfLINE_NUM_COEFF[14],sRPC.adfLINE_NUM_COEFF[15],sRPC.adfLINE_NUM_COEFF[16],sRPC.adfLINE_NUM_COEFF[17],sRPC.adfLINE_NUM_COEFF[18],sRPC.adfLINE_NUM_COEFF[19],
			sRPC.adfLINE_DEN_COEFF[0],sRPC.adfLINE_DEN_COEFF[1],sRPC.adfLINE_DEN_COEFF[2],sRPC.adfLINE_DEN_COEFF[3],sRPC.adfLINE_DEN_COEFF[4],sRPC.adfLINE_DEN_COEFF[5],sRPC.adfLINE_DEN_COEFF[6],sRPC.adfLINE_DEN_COEFF[7],sRPC.adfLINE_DEN_COEFF[8],sRPC.adfLINE_DEN_COEFF[9],
			sRPC.adfLINE_DEN_COEFF[10],sRPC.adfLINE_DEN_COEFF[11],sRPC.adfLINE_DEN_COEFF[12],sRPC.adfLINE_DEN_COEFF[13],sRPC.adfLINE_DEN_COEFF[14],sRPC.adfLINE_DEN_COEFF[15],sRPC.adfLINE_DEN_COEFF[16],sRPC.adfLINE_DEN_COEFF[17],sRPC.adfLINE_DEN_COEFF[18],sRPC.adfLINE_DEN_COEFF[19],
			sRPC.adfSAMP_NUM_COEFF[0],sRPC.adfSAMP_NUM_COEFF[1],sRPC.adfSAMP_NUM_COEFF[2],sRPC.adfSAMP_NUM_COEFF[3],sRPC.adfSAMP_NUM_COEFF[4],sRPC.adfSAMP_NUM_COEFF[5],sRPC.adfSAMP_NUM_COEFF[6],sRPC.adfSAMP_NUM_COEFF[7],sRPC.adfSAMP_NUM_COEFF[8],sRPC.adfSAMP_NUM_COEFF[9],
			sRPC.adfSAMP_NUM_COEFF[10],sRPC.adfSAMP_NUM_COEFF[11],sRPC.adfSAMP_NUM_COEFF[12],sRPC.adfSAMP_NUM_COEFF[13],sRPC.adfSAMP_NUM_COEFF[14],sRPC.adfSAMP_NUM_COEFF[15],sRPC.adfSAMP_NUM_COEFF[16],sRPC.adfSAMP_NUM_COEFF[17],sRPC.adfSAMP_NUM_COEFF[18],sRPC.adfSAMP_NUM_COEFF[19],
			sRPC.adfSAMP_DEN_COEFF[0],sRPC.adfSAMP_DEN_COEFF[1],sRPC.adfSAMP_DEN_COEFF[2],sRPC.adfSAMP_DEN_COEFF[3],sRPC.adfSAMP_DEN_COEFF[4],sRPC.adfSAMP_DEN_COEFF[5],sRPC.adfSAMP_DEN_COEFF[6],sRPC.adfSAMP_DEN_COEFF[7],sRPC.adfSAMP_DEN_COEFF[8],sRPC.adfSAMP_DEN_COEFF[9],
			sRPC.adfSAMP_DEN_COEFF[10],sRPC.adfSAMP_DEN_COEFF[11],sRPC.adfSAMP_DEN_COEFF[12],sRPC.adfSAMP_DEN_COEFF[13],sRPC.adfSAMP_DEN_COEFF[14],sRPC.adfSAMP_DEN_COEFF[15],sRPC.adfSAMP_DEN_COEFF[16],sRPC.adfSAMP_DEN_COEFF[17],sRPC.adfSAMP_DEN_COEFF[18],sRPC.adfSAMP_DEN_COEFF[19],
			sRPC.dfMIN_LONG, sRPC.dfMIN_LAT, sRPC.dfMAX_LONG, sRPC.dfMAX_LAT,std::max<double>(min_pixel-offset_buffer, 0),std::max<double>(min_line-offset_buffer, 0));
	fclose(fid);
	
	return ret;
}



int main(int argc, char* argv[]) {
  try {
    TCLAP::CmdLine cmd("crop-nif command line", ' ', "1.0");

    TCLAP::UnlabeledValueArg<std::string> input_filename_arg("input", "image file", true,  "", "image file");
    TCLAP::ValueArg<std::string> kml_filename_arg("k", "kml", "KML filename", true, "", "kml file");
    TCLAP::ValueArg<std::string> output_filename_arg("o", "output", "output filename (extension determines type, jpg, tiff, and laz are supported)", true, "", "output file");
    TCLAP::ValueArg<std::string> rpc_filename_arg("r", "rpc", "rpc output filename", true, "", "rpc file");
     TCLAP::ValueArg<double> spacing_arg("s", "spacing", "output ground spatial density (applies to las/laz files only)", false, 1,
    									"ground spatial density");
    cmd.add(input_filename_arg);
    cmd.add(kml_filename_arg);
    cmd.add(output_filename_arg);
    cmd.add(rpc_filename_arg);
    cmd.add(spacing_arg);
    cmd.parse(argc, argv);

    std::string input_file = input_filename_arg.getValue();
    std::string kml_file = kml_filename_arg.getValue();
    std::string output_file = output_filename_arg.getValue();
    std::string rpc_file = rpc_filename_arg.getValue();
    
	CropImageWithKML(input_file, kml_file, rpc_file, output_file, spacing_arg.getValue());

  } catch (std::exception& e) {
    std::cerr << "error: " << e.what() << std::endl;
    return EXIT_FAILURE;
  }

  return EXIT_SUCCESS;
}
